# Create XML to invoke a SOAP method

"writeTypes" =
function(x, con, types = getSOAPType(x))
{  
    if(!is.null(types)) {

      if(is(types, "ArrayType")) {
        tmp = paste(getSOAPType(types), collapse = " ")
#        tmp = paste("xsi:type=\"SOAP-ENC:Array\" SOAP-ENC:arrayType =\"",  getSOAPType(types@elType), "\"", collapse = "")
      } else {
        tmp = paste(names(types), paste("\"", types, "\"", sep=""), sep="=", collapse= " ")
      }
      cat(tmp, file=con)       
    }

    tmp
}


setGeneric("toSOAP", function(obj, con = xmlOutputBuffer(header=""), type = NULL, literal = FALSE, ...) standardGeneric("toSOAP"))

setMethod("toSOAP", c("vector", type = "NULL"),
              function(obj, con = xmlOutputBuffer(header=""), type = NULL, literal = FALSE, ...)
                   toSOAPArray(obj, con, type = type, literal = literal, ...))

setMethod("toSOAP", c("list", type="NULL"),
function(obj, con = xmlOutputBuffer(header=""), type = NULL, literal = FALSE, ...)
{
 if(length(names(obj)))
   return(toSOAPNamedContainer(obj, con, type, ...))
          
 if(length(obj) > 1) {
    # If we have named elements, then we need to do something with these rather than lose them
    # so we put them out as named elements, i.e. as a struct.
  if(!is.null(names(obj))) {
    for(i in names(obj)) {
      cat("<", i, file = con, sep="")
      writeTypes(obj[[i]])
      cat(obj[[i]])
    }
  } else {
    toSOAPArray(obj, con, type = type, literal = literal, ...)
  }
 } else {
    cat(obj, file=con)
 }

 invisible(TRUE)
})


setMethod("toSOAP", c("ANY", type = "ClassDefinition"),
function(obj, con = xmlOutputBuffer(header=""), type = NULL, literal = FALSE, ...)
{
  if(literal) {
     for(i in names(getSlots(class(obj)))) {
       cat("<", i, ">", sep = "", file = con)
       toSOAP(slot(obj, i), con = con, type = type@slotTypes[[i]], literal = literal)
       cat("</", i, ">\n", sep = "", file = con)       
     }
   } else {

     stop("No code yet or toSOAP for object and ClassDefinition with literal = FALSE")
   }

   invisible(TRUE)
})


setGeneric("toSOAPArray", function(obj, con = stdout(), type = NULL, literal = FALSE, ...)
                                     standardGeneric("toSOAPArray"))

setMethod("toSOAPArray", "vector", 
function(obj, con = stdout(), type = NULL, literal = FALSE, ...)
{
   if(length(names(obj))) {
     return(toSOAPNamedContainer(obj))
   }
  
   if(length(obj) == 1 && is.null(type)) {
      cat(obj, file = con)
      return(TRUE)
   }

   newLine = "\r\n"
   newLine=""   

   if(!is.null(type))
     elType = type@elType
   else
     elType = NULL


    for(i in obj) {
if(TRUE) {         
      cat("<item ", file=con)
      writeTypes(i, con) # , type = elType)
      cat(">", newLine, sep="", file=con)
      cat(i, file=con)
      cat(newLine, "</item>", sep="", file=con)
} else {
  tmp = strsplit(getSOAPType(typeof(i)), ":")[[1]][2]
  cat("<", tmp, ">", sep="", file = con)
  cat(i, file=con)
  cat("</", tmp, ">", sep="", file=con)
}
    }
#   cat("</SOAP-ENC:Array>", file = con)
})


SOAPbase64Decode =
function(x) 
  base64Decode(xmlValue(x))




# Named list of mappings from primitive SOAP types
# to S data types.
# From http://www.w3.org/TR/xmlschema-2/#built-in-primitive-datatypes
# d means done
# w means "needs more work"
##  d    3.2.1 string
##  d    3.2.2 boolean
##  d    3.2.3 decimal
##  d    3.2.4 float         
##  d    3.2.5 double
##       3.2.6 duration
##  d    3.2.7 dateTime
##  d    3.2.8 time
##  d    3.2.9 date
##       3.2.10 gYearMonth
##       3.2.11 gYear
##       3.2.12 gMonthDay
##       3.2.13 gDay
##       3.2.14 gMonth
##  w    3.2.15 hexBinary
##  w    3.2.16 base64Binary
##       3.2.17 anyURI
##       3.2.18 QName
##       3.2.19 NOTATION

# Is there an int? Yes, it is restricted from other types.
# See the tree at the top of section 3.2 of the document above.

SOAPTypes <-
  list("character" = list("xsi:type" = "xsd:string", type = "string"),
       "numeric"   = list("xsi:type" = "xsd:float", type = "float"),
       "numeric"   = list("xsi:type" = "xsd:long", type = "long"),
       "numeric"   = list("xsi:type" = "xsd:decimal", type = "decimal"),
#?      "double"   = list("xsi:type" = "xsd:float", type = "float"),
       "numeric"   = list("xsi:type" = "xsd:double", type = "double"),       
       "integer"   = list("xsi:type" = "xsd:int", type = "int"),
       "logical"   = list("xsi:type" = "xsd:boolean", type = "boolean"),

       "POSIXct"   = list("xsi:type" = "xsd:date", type = "date"),
       "POSIXct"   = list("xsi:type" = "xsd:dateTime", type = "dateTime"),
       "POSIXct"   = list("xsi:type" = "xsd:time", type = "time"),

       "raw"       = list("xsi:type" = "xsd:base64Binary", type = "base64Binary", from = SOAPbase64Decode),
       "raw"       = list("xsi:type" = "xsd:hexBinary", type = "hexBinary"),

       "parseURI"       = list("xsi:type" = "xsd:anyURI", type = "anyURI"),
#QName
#NOTATION
       "NULL"      = list("xsi:null" = 1, "xsi:type" = "null", type = "null")
    )


setMethod("toSOAP", c(type = "ArrayType"),
              function(obj, con = xmlOutputBuffer(header=""), type = NULL, literal = FALSE, ...) {
                toSOAPArray(obj, con = con, type = type, literal = literal, ...)
              })


setMethod("toSOAP", c("vector", type = "PrimitiveSOAPType"),
              function(obj, con = xmlOutputBuffer(header=""), type = NULL, literal = FALSE, ...) {
                val = type@toConverter(obj)
                if(length(val) > 1)
                  warning("Converting value to primitive SOAP type results in vector with more than one element. Ignoring remainder.")
                cat(val[1], file = con)
              }
)


setMethod("toSOAP", c("vector", type = "SOAPType"),
              function(obj, con = xmlOutputBuffer(header=""), type = NULL,  literal = FALSE, ...) {
                toSOAPArray(obj, con, type = type, literal = literal, ...)
              })



setGeneric("getSOAPType", function(obj, value = NULL) standardGeneric("getSOAPType"))

setMethod("getSOAPType", "ANY",
function(obj, value = NULL)
{
  if(is.null(obj)) #XXX || length(obj) == 0)
    return(SOAPTypes[["NULL"]][["xsi:type"]])

   ans <- character(0)
  
  if(length(obj) > 1) {
         # Array
    n <- length(obj)
    same <- sapply(obj, function(x, target) typeof(x) == target, target = typeof(obj[[1]]))

     ans <- c("xsi:type"="SOAP-ENC:Array")
    if(all(same)) {
      type <- getSOAPType(obj[[1]])
      ans["SOAP-ENC:arrayType"]  = paste(type[1], "[", n, "]", sep="")
    }
  } else {
     m <- typeof(obj)
     ans["xsi:type"] <- SOAPTypes[[m]][["xsi:type"]]
  }
  
  return(ans)  
})

setMethod("getSOAPType", "PrimitiveSOAPType",
function(obj, value = NULL) {
  c("xsi:type" = paste(obj@ns, obj@name, sep=":"))
})


setMethod("getSOAPType", "ArrayType",
function(obj, value = NULL) {
    ans <- c("xsi:type"="SOAP-ENC:Array")

    if(!is.null(value))  {
      n = length(value)
      type <- getSOAPType(value[1])
    } else {
       n = ""
       type="ur-type"
    }
    ans["SOAP-ENC:arrayType"]  = paste(type[1], "[", n, "]", sep="")

    ans
})


.SOAPAction <-
function(action, method, server, xmlns, sep="#")
{
  if(missing(sep)) {
    if(substring(action, nchar(action)) == "#" || substring(method, 1, 1) == "#")
      sep=""
  }

  if(action == "")
    action = xmlns
  
  paste(action, sep, method, sep="")
}

.SOAPDefaultHandlers <-
  list(action = .SOAPAction,
       result = function(xmlSource, header, method, server) {
         response <- parseSOAP(xmlSource, asText = TRUE)
         fromSOAP(response[[1]])
       })

SOAPHandlers <-
function(..., include = character(0), exclude = character(0))
{  
 defaults <- .SOAPDefaultHandlers
 els <- list(...)

 .merge(els, defaults, include, exclude)
}



SOAPResult =
function(content, header, obj = new("SOAPResult"))
{
 obj@content = content
 obj@header = as.list(header)
 obj
}


  


writeSOAPMessage <-
function(con, nameSpaces, method, ...,
          .types = NULL, xmlns = NULL, .soapArgs = list(),
          .literal = FALSE)
{
  #XXX make this anonymous!
 con <- textConnection(".foo", open="w", local = TRUE)
 writeSOAPEnvelope(con, nameSpaces = nameSpaces)

 writeSOAPBody(method, ..., .soapArgs = .soapArgs, xmlns = xmlns, con = con,
                .types = .types, .literal = .literal)

 cat("</SOAP-ENV:Envelope>\r\n", file=con)

 cat("\r\n", file=con)
 close(con)
   # Can we just access .foo here directly.
 paste(get(".foo"), collapse="\n")
}  


#SOAP 1.1
#      "SOAP-ENC" = 'http://schemas.xmlsoap.org/soap/encoding/',
#      "xsd" = 'http://www.w3.org/1999/XMLSchema',
# SOAP  1.2 
#      "SOAP-ENV" = 'http://www.w3.org/2001/06/soap-envelope',
#      "SOAP-ENC" = 'http://www.w3.org/2001/06/soap-encoding',
#      "xsd" = 'http://www.w3.org/2001/XMLSchema',

.SOAPDefaultNameSpaces <-
  list("1.1" = 
         c(
           'SOAP-ENC'="http://schemas.xmlsoap.org/soap/encoding/",
           'SOAP-ENV'="http://schemas.xmlsoap.org/soap/envelope/",           
           'xsi'="http://www.w3.org/1999/XMLSchema-instance",
           'xsd'="http://www.w3.org/1999/XMLSchema"
           )  ,
       "1.2" = 
         c(
           'SOAP-ENC'="http://www.w3.org/soap-encoding/",
           'SOAP-ENV'="http://schemas.xmlsoap.org/soap/envelope/",                      
#XXX       'SOAP-ENV'="http://www.w3.org/2001/06/soap-envelope",
           'xsi'="http://www.w3.org/2001/XMLSchema-instance",
           'xsd'="http://www.w3.org/2001/XMLSchema"
           )  )       
       

.merge <-
function(els, defaults, include = NULL, exclude = NULL)
{
 if(length(els) > 0) {
   which <- match(names(defaults), names(els))
   if(any(!is.na(which)))
     els[names(defaults)[!is.na(which)]] <- defaults[!is.na(which)]
 } else
   els <- defaults

 if(length(include)) {
   els <- els[include]
 } else if(length(exclude)) {
   which <- match(exclude, names(els))
   if(any(!is.na(which)))
     els <- els[- (which[!is.na(which)])]
 }

 els   
}  

SOAPNameSpaces <-
function(..., include = character(0), exclude = character(0), version = getOption("SSOAP:DefaultNamespace"))
{
  # If version doesn't match an entry, then use NULL rather than raising an error.
  #  if(!(version %in% names(.SOAPDefaultNameSpaces))) 
  #     stop("Incorrect SOAP name space version: ", version)

  defaults <- .SOAPDefaultNameSpaces[[version]]
  els <- sapply(list(...), as.character)
  .merge(els, defaults, include, exclude)
}

  # This and writeSOAPBody should use the XMLOutput.. classes in the XML package.
writeSOAPEnvelope <-
function(con, nameSpaces = .SOAPNameSpaces())
{
  cat("<?xml version='1.0' encoding='UTF-8'?>\r\n", file = con)
  cat("<SOAP-ENV:Envelope\r\n", file=con)
  for(i in names(nameSpaces)) {
     cat("xmlns:", i, "=\"", nameSpaces[i],"\"\r\n", sep="", file=con)
  }
  cat('SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"', file=con)
  cat(">\r\n", file=con)
}


writeSOAPBody <-
function(method, ..., xmlns= NULL, con,
          .types = NULL, .soapArgs = list(),
          .literal = FALSE)
{
  # Write the preamble for the SOAP body.
 cat('<SOAP-ENV:Body>\r\n', file=con)

 nsName <- ""

 if(!is.null(xmlns) && !is.null(names(xmlns))) 
   nsName <- paste(names(xmlns)[1], ":", sep="")


  # Output the namespace and the method name
 cat("<", nsName, method, sep="", file=con)  

  # Now announce the namespace attribute
 if(!is.null(xmlns) && xmlns != "") {
   ns <- ""
   if(!is.null(names(xmlns)))
     ns <-  paste(":", names(xmlns)[1], sep="")
   cat(" xmlns", ns,"=\"", xmlns,"\"", file=con, sep="")
 }
 cat(">\n", file=con)

   # Now we are moving onto the parameters for the call.
 args <- list(...)
 args <- merge.list(args, .soapArgs)
 
 argNames <- names(args)

   # Patch up argument names if necessary.
   # If none, make them up as a1, a2, a3,
 if(is.null(argNames)) {
   argNames <- paste("arg", 1:length(args), sep="")
 } else {
     # If there are any missing, add them as arg1, arg2, ...
   which = argNames == ""
   if(any(which))
     argNames[which] <- paste("arg", 1:sum(which), sep="")
 }

 
  # Now write the arguments into the body.
  # Each argument looks like <argName type>value</argName>
 if(length(args) > 0) {
  for(i in 1:length(args)) {
    argName <- argNames[i]

    typedef = NULL
    if(length(.types) >= i ) {
      typedef = .types[[i]]
      type = getSOAPType(.types[[i]], args[[i]])
    } else
      type = getSOAPType(args[[i]])


    if(!.literal) {
       cat("<", argName, " ", sep="", file=con)
       if(length(type) && type[1] != "")
          writeTypes(args[[i]], con = con, types = type)

       cat(">", file=con)
    }
    toSOAP(args[[i]], con, type = typedef, literal = .literal)
    if(!.literal)
       cat("</", argName, ">\r\n", sep="", file=con)  
  }
 }
 cat("</", nsName, method, ">\r\n", sep="", file=con) 

 cat("</SOAP-ENV:Body>\r\n", file=con) 
}  


toSOAPNamedContainer <-
function(obj, con = stdout(), type = NULL, ..., .literal = FALSE)
{
   # cat("xsi:type=\"apachens:Map\">", , file = con,)
   for(i in names(obj)) {
     cat("<item><key xsi:type=\"xsd:string\">", i, "</key>", file = con, sep = "")
     cat("<value", file = con)
     toSOAP(obj[[i]],  con = con, type = NULL, .literal = .literal)
     cat("</value>", file = con)     
   }

   return(TRUE)
}


 
