.SOAP <-
function(server, method, ..., .soapArgs = list(), action, nameSpaces = SOAPNameSpaces(), xmlns = NULL,
         handlers = SOAPHandlers(), .types = NULL, .convert = TRUE,
         .opts = list(), curlHandle = getCurlHandle(),
         .header = getSOAPRequestHeader(action, .server = server),
         .literal = FALSE)  
{
 library(RCurl)

 if(is.null(xmlns))
   xmlns <- c(namesp1 = action)

 if(!is(action, "AsIs")) {
   if((!is(action, c("SOAPAction")) || action == "") && !is.null(handlers) && "action" %in% names(handlers)) 
      action <- handlers[["action"]](action, method, server, xmlns)
 }

 txt <- writeSOAPMessage(con, nameSpaces, method, ...,
                          .soapArgs = .soapArgs, .types = .types, xmlns = xmlns, .literal = .literal)

# Curl specific
 headerDataFun = basicTextGatherer()
 bodyDataFun = basicTextGatherer()  

 if(is(server, "CURLHandle") && missing(curlHandle) )
   curlHandle = server
 else 
   .opts$url = toURL(server)

 if(length(.header))
   .opts$httpheader = .header

 curlPerform(postfields = txt,
             writeFunction = bodyDataFun$update,
             headerFunction = headerDataFun$update, 
             .opts = .opts, curl = curlHandle)

# end of Curl specific part.
 
 content = structure(
              list(header = getHTTPResponseHeader(headerDataFun$value(NULL)),
                   content = bodyDataFun$value()),
              class = "SOAPHTTPReply"
              )

 if(isHTTPError(content$header)) {
     # This would be the exception that we throw if we had an exception system.
   fault <- SOAPFault(parseSOAP(content$content, asText = TRUE))
    # 
   stop("Error occurred in the HTTP request: ", fault@message, fault@detail)
   return(fault)
 }


 if(is.logical(.convert) && .convert && !is.null(handlers) && !is.na(match("result", names(handlers))))
   return(handlers[["result"]](content$content,  content$header, method))
 else if(is.function(.convert))
   return(.convert(content))
 else if(is(.convert, "SOAPType")) 
   return(convertFromSOAP(SOAPResult(content$content, content$header), .convert))

 return(content) 
}

toURL =
function(server)
{
 if(is.character(server))
   return(server)

 port = character(0)
 if(!is.na(server@port) )
   port = as.character(server@port)
 
 url = paste(getProtocol(server), server@host, 
              ifelse(length(port), ":", ""), port,
              ifelse(substring(server@url, 1, 1) == "/", "", "/"),
              server@url, sep="")
 url
}


getProtocol = 
function(server)
{
  if(is(server, "HTTPSSOAPServer"))
    return("https://")
  else if(is(server, "HTTPSOAPServer"))
     return("http://")

  standardGeneric("server")
}

getSOAPRequestHeader =
function(action, ...,
         defaultHeader = c(Accept="text/xml", Accept="multipart/*",
                           'Content-Type' = "text/xml; charset=utf-8"),
         .server = NULL)
{

   httpHeader= defaultHeader
     # Put the action in quotes.
   httpHeader["SOAPAction"] = paste('"', action, '"', sep="")

   httpHeader
}


getHTTPResponseHeader <-
  #
  # Reads the lines from the HTTP response 
  # header and converts them into a
  # name=value  collection as a character vector.
  #
function(lines)
{
 if(length(lines) < 1)
   return(NULL)

 status = lines[1]
 lines = lines[-c(1, length(lines))]
 lines = gsub("\r\n\0$","", lines)

 if(FALSE) { # Doesn't behave.
   header = lines[-1]
   header <- read.dcf(textConnection(header))
 } else {
   els <- sapply(lines, function(x) strsplit(x, ":"))
   header <- sapply(els, function(x) x[2])
   names(header) <- sapply(els, function(x) x[1]) 
 }

 els <- strsplit(status, " ")[[1]] 
 header[["status"]] <- as.integer(els[2])
 header[["statusMessage"]] <- els[3]
 
 header
}


.onLoad <-
function(lib, pkg)
{
 options("SSOAP:DefaultNamespace" = "1.2")
}
